//Edit various PlayerPawn functions

//************************************
//Precached variables go here

static INT NAMES_PlayerPawn_Funcs = 0;
static FName NAME_PP_CanSpectate;
static FName NAME_PP_BecomeViewTarget;
static FName NAME_PP_Say;
static FName NAME_PP_CallForHelp;
static FName NAME_PP_AllowsBroadcast;
static FName NAME_PP_MutatorTeamMessage;
static FName NAME_PP_GetHumanName;


//************************************
//Definitions, classes and events go here

struct APlayerPawn_Dummy : public APlayerPawn
{
	// UTPG --
	INT LoginAttempts;		// How many login attempts we've made
	UBOOL bLoginDisabled;	// no longer able to login even if correct password
	FLOAT NextLoginTime;	// Time at which game will accept another attempt to login

	// All received from GameInfo
	FLOAT ViewDelay;		// Seconds between each viewplayer()/viewplayernum()
	FLOAT TauntDelay;		// Seconds between taunts
	FLOAT SpeechDelay;		// Seconds between chat messages

	FLOAT LastView;			// Used to track last view
	FLOAT LastTaunt;		// Used to track last taunt
	FLOAT LastSpeech;		// Used to track last chat message

	UBOOL bCyclingView;		// Used to allow ViewClass to disregard LastView

	FLOAT MinFOV;			// Enforced MinFOV for this server
	FLOAT MaxFOV;			// Enforced MaxFOV for this server

	INT MaxNameChanges;		// Max number of time a player can change their name in the game
	INT NameChanges;		// Used to track name changes

	void ViewClass( UClass* aClass, UBOOL bQuiet=0);
	void ViewPlayer( const FString& S);
	void TeamSay( FString& Msg);
	DECLARE_FUNCTION( execViewClass);
	DECLARE_FUNCTION( execViewPlayerNum);
	DECLARE_FUNCTION( execViewPlayer);
	DECLARE_FUNCTION( execTeamSay);
};

/*
struct APlayerPawn_eventSendVoiceMessageParams
{
	APlayerReplicationInfo* Sender;
	APlayerReplicationInfo* Recipient;
	FName MessageType;
	BYTE MessageID;
	FName BroadcastType;
};*/

//************************************
//Bind function to native, register uscript function names

static void ReplacePlayerPawnFuncs( UBOOL bSet)
{
	guard(ReplacePlayerPawnFuncs);

	UClass* PlayerPawn = APlayerPawn::StaticClass();
	
	UFunction* ViewPlayerNum = FindBaseFunction( PlayerPawn, TEXT("ViewPlayerNum"));
	UFunction* ViewClass = FindBaseFunction( PlayerPawn, TEXT("ViewClass"));
	UFunction* ViewPlayer = FindBaseFunction( PlayerPawn, TEXT("ViewPlayer"));
	UFunction* TeamSay = FindBaseFunction( PlayerPawn, TEXT("TeamSay"));
	if ( !ViewPlayerNum || !ViewClass || !ViewPlayer || !TeamSay ) //Script not serialized?
		return;

	if ( !bSet )
	{
		ViewPlayerNum->Func = &UObject::ProcessInternal;
		ViewPlayerNum->FunctionFlags &= ~FUNC_Native;
		ViewClass->Func = &UObject::ProcessInternal;
		ViewClass->FunctionFlags &= ~FUNC_Native;
		ViewPlayer->Func = &UObject::ProcessInternal;
		ViewPlayer->FunctionFlags &= ~FUNC_Native;
		TeamSay->Func = &UObject::ProcessInternal;
		TeamSay->FunctionFlags &= ~FUNC_Native;
	}
	else
	{
		ViewPlayerNum->Func = (Native)&APlayerPawn_Dummy::execViewPlayerNum;
		ViewPlayerNum->FunctionFlags |= FUNC_Native;
		ViewClass->Func = (Native)&APlayerPawn_Dummy::execViewClass;
		ViewClass->FunctionFlags |= FUNC_Native;
		ViewPlayer->Func = (Native)&APlayerPawn_Dummy::execViewPlayer;
		ViewPlayer->FunctionFlags |= FUNC_Native;
		TeamSay->Func = (Native)&APlayerPawn_Dummy::execTeamSay;
		TeamSay->FunctionFlags |= FUNC_Native;
	}
	
	//Register subnames
	if ( bSet && !NAMES_PlayerPawn_Funcs )
	{
		NAMES_PlayerPawn_Funcs = 1;
		NAME_PP_CanSpectate			= FName(TEXT("CanSpectate"), FNAME_Intrinsic);
		NAME_PP_BecomeViewTarget	= FName(TEXT("BecomeViewTarget"), FNAME_Intrinsic);
		NAME_PP_Say					= FName(TEXT("Say"), FNAME_Intrinsic);
		NAME_PP_CallForHelp			= FName(TEXT("CallForHelp"), FNAME_Intrinsic);
		NAME_PP_AllowsBroadcast		= FName(TEXT("AllowsBroadcast"), FNAME_Intrinsic);
		NAME_PP_MutatorTeamMessage	= FName(TEXT("MutatorTeamMessage"), FNAME_Intrinsic);
		NAME_PP_GetHumanName		= FName(TEXT("GetHumanName"), FNAME_Intrinsic);
	}

	unguard;
}

//************************************
//Utilitary/common methods used by natives

static void PlayerPawnSay( APlayerPawn* PP, FString& Msg)
{
	if ( !PP )
		return;
	INT Params[3];
	for ( INT i=0 ; i<3 ; i++ ) //SUPER HACK, DO NOT ALLOCATE MEMORY
		Params[i] = ((INT*)&Msg)[i];
	UFunction* Say = PP->FindFunction( NAME_PP_Say);
	if ( Say )
		PP->ProcessEvent( Say, Params);
}

static UBOOL GameCanSpectate( AGameInfo* Game, APawn* Viewer, AActor* ViewTarget)
{
	if ( !Game )
		return 0;
	INT Params[3];
	Params[0] = (INT) Viewer;
	Params[1] = (INT) ViewTarget;
	Params[2] = 0; //Return value
	UFunction* CanSpectate = Game->FindFunction( NAME_PP_CanSpectate);
	if ( CanSpectate )
		Game->ProcessEvent( CanSpectate, Params);
	return Params[2];
}

static UBOOL GameAllowsBroadcast( AGameInfo* Game, AActor* Broadcaster, INT Len )
{
	if ( !Game )
		return 1;
	INT Params[3];
	Params[0] = (INT) Broadcaster;
	Params[1] = Len;
	Params[2] = 0; //Return value
	UFunction* AllowsBroadcast = Game->FindFunction( NAME_PP_AllowsBroadcast);
	if ( AllowsBroadcast )
		Game->ProcessEvent( AllowsBroadcast, Params);
	return Params[2];
}

static FString ActorGetHumanName( AActor* Actor)
{
	FString Result(0);
	if ( Actor )
	{
		UFunction* GetHumanName = Actor->FindFunction( NAME_PP_GetHumanName);
		if ( GetHumanName )
			Actor->ProcessEvent( GetHumanName, &Result);
	}
	return Result;
}

static UBOOL MutatorMutatorTeamMessage( AMutator* Mut, AActor* Sender, APawn* Receiver, APlayerReplicationInfo* PRI, FString& S, FName Type, BITFIELD bBeep )
{
	if ( !Mut )
		return 1;
	INT Params[9];
	Params[0] = (INT) Sender;
	Params[1] = (INT) Receiver;
	Params[2] = (INT) PRI;
	Params[3] = ((INT*)&S)[0];
	Params[4] = ((INT*)&S)[1];
	Params[5] = ((INT*)&S)[2];
	Params[6] = ((INT*)&Type)[0];
	Params[7] = bBeep;
	Params[8] = 0; //Return value
	UFunction* MutatorTeamMessage = Mut->FindFunction( NAME_PP_MutatorTeamMessage);
	if ( MutatorTeamMessage )
		Mut->ProcessEvent( MutatorTeamMessage, Params);
	return Params[8];
}

static void ActorBecomeViewTarget( AActor* Other)
{
	if ( !Other )
		return;
	UFunction* BecomeViewTarget = Other->FindFunction(NAME_PP_BecomeViewTarget);
	if ( BecomeViewTarget )
		Other->ProcessEvent( BecomeViewTarget, NULL);
}

static FString GetBestName( AActor* Other)
{
	if ( !Other )
		return FString(0);
	APawn* P = Cast<APawn>(Other);
	if ( P && P->PlayerReplicationInfo && P->PlayerReplicationInfo->PlayerName.Len() )
		return P->PlayerReplicationInfo->PlayerName;
	return Other->GetName();
}

void APlayerPawn_Dummy::ViewClass( UClass* aClass, UBOOL bQuiet )
{
	guard( ViewClass);
	if ( !Level || !XLevel || !aClass )
		return;
	if ( Level->Game && !Level->Game->bCanViewOthers )
		return;
	
	if ( XCGE_Defaults->b451Setup && !bCyclingView )
	{
		if ( (ViewDelay > 0.0f) && (Level->TimeSeconds - LastView < ViewDelay) )
			return;
		LastView = Level->TimeSeconds;
	}

	//This optimizes the following loop
	if ( ViewTarget && !ViewTarget->IsA(aClass) )
		ViewTarget = NULL;
	
	AActor* First = NULL;
	UBOOL bFound = 0;
	INT i=0;
	for ( ; i<XLevel->Actors.Num() ; i++ )
	{
		AActor* Other = XLevel->Actors(i);
		if ( Other && !Other->bDeleteMe && Other->IsA( aClass) )
		{
			if ( !First && (Other != this) && ((bAdmin && !Level->Game) || GameCanSpectate(Level->Game, this, Other))  )
			{
				First = Other;
				if ( (bFound == 2) || !ViewTarget ) //No need to keep querying
					break;
				bFound = 1;
			}
			if ( Other == ViewTarget )
			{
				First = NULL;
				bFound = 2;
			}
		}
	}
	if ( First )
	{
		if ( !bQuiet )
			eventClientMessage( ViewingFrom + GetBestName(First), NAME_Event, 1);
	}
	else if ( !bQuiet )
	{
		if ( bFound )
			eventClientMessage( ViewingFrom + OwnCamera, NAME_Event, 1);
		else
			eventClientMessage( FailedView, NAME_Event, 1);
	}
	ViewTarget = First;

	bBehindView = (ViewTarget != NULL);
	if ( bBehindView )
		ActorBecomeViewTarget(ViewTarget);
	unguard;
}


void APlayerPawn_Dummy::ViewPlayer( const FString& S )
{
	guard(ViewPlayer);
	if ( !Level || !S.Len() )
		return;

	APawn* Best = NULL;
	for ( APawn* P=Level->PawnList; P; P=P->nextPawn )
		if ( P->bIsPlayer && P->PlayerReplicationInfo )
		{
			if ( P->PlayerReplicationInfo->PlayerName == S )
			{
				Best = P;
				break;
			}
			else if ( P->PlayerReplicationInfo->PlayerName.InStr(S) >= 0 )
				Best = P;
		}

	if ( Best && GameCanSpectate(Level->Game, this, Best) )
	{
		eventClientMessage( ViewingFrom + TEXT(" ") + Best->PlayerReplicationInfo->PlayerName, NAME_Event, 1);
		ViewTarget = (Best==this) ? NULL : Best;
	}
	else
		eventClientMessage(FailedView, NAME_Event, 0);

	bBehindView = (ViewTarget != NULL);
	if ( bBehindView )
		ActorBecomeViewTarget(ViewTarget);
	unguard;
}


void APlayerPawn_Dummy::TeamSay( FString& Msg)
{
	if ( !Level || !PlayerReplicationInfo )
		return;

	if ( !Level->Game || !Level->Game->bTeamGame )
	{
		PlayerPawnSay( this, Msg);
		return;
	}
	
	if ( Msg == TEXT("help") )
	{
		UFunction* CallForHelp = FindFunction( NAME_PP_CallForHelp);
		if ( CallForHelp )
			ProcessEvent( CallForHelp, NULL);
		return;
	}

	FString LeftChop(0);
	FString RightChop(0);
	FString Processed( Msg);
	//Chop only once each parameter
	if ( Processed.Split( FString( TEXT("%H")), &LeftChop, &RightChop) )
		Processed = FString::Printf( TEXT("%s%i health%s"), *LeftChop, Health, *RightChop);
	if ( Processed.Split( FString( TEXT("%h")), &LeftChop, &RightChop) )
		Processed = FString::Printf( TEXT("%s%ihp%s"), *LeftChop, Health, *RightChop);
	if ( Processed.Split( FString( TEXT("%A")), &LeftChop, &RightChop) )
	{
		INT Armor = 0;
		for ( AInventory* Inv=Inventory ; Inv ; Inv=Inv->Inventory )
			if ( Inv->bIsAnArmor ) Armor += Inv->Charge;
		Processed = FString::Printf( TEXT("%s%i armor %s"), *LeftChop, Armor, *RightChop);
	}
	if ( Processed.Split( FString( TEXT("%a")), &LeftChop, &RightChop) )
	{
		INT Armor = 0;
		for ( AInventory* Inv=Inventory ; Inv ; Inv=Inv->Inventory )
			if ( Inv->bIsAnArmor ) Armor += Inv->Charge;
		Processed = FString::Printf( TEXT("%s%ia %s"), *LeftChop, Armor, *RightChop);
	}
	if ( Processed.Split( FString( TEXT("%Z")), &LeftChop, &RightChop) || Processed.Split( FString( TEXT("%z")), &LeftChop, &RightChop) )
	{
		FString ZoneText(0);
		if ( !Region.Zone )
			ZoneText = TEXT("(undefined)");
		else
		{
			if ( PlayerReplicationInfo && PlayerReplicationInfo->PlayerLocation )
				ZoneText = PlayerReplicationInfo->PlayerLocation->LocationName;
			else if ( Region.Zone->ZoneName.Len() )
				ZoneText = Region.Zone->ZoneName;
			if ( !ZoneText.Len() )
				ZoneText = FString::Printf(TEXT("Zone [%i]"), Region.ZoneNumber);
		}
		Processed = FString::Printf( TEXT("%s%s %s"), *LeftChop, *ZoneText, *RightChop);
	}
	if ( Processed.Split( FString( TEXT("%W")), &LeftChop, &RightChop) || Processed.Split( FString( TEXT("%w")), &LeftChop, &RightChop) )
		Processed = FString::Printf( TEXT("%s%s%s"), *LeftChop, Weapon ? *ActorGetHumanName(Weapon) : TEXT("[]"),*RightChop);

	if ( GameAllowsBroadcast( Level->Game, this, Processed.Len()) )
	{
		FName TeamSayName = FName( TEXT("TeamSay")); //Cheap
		for ( APawn* P=Level->PawnList ; P ; P=P->nextPawn )
			if ( P->bIsPlayer && P->PlayerReplicationInfo && (P->PlayerReplicationInfo->Team == PlayerReplicationInfo->Team) && Cast<APlayerPawn>(P) )
				if ( MutatorMutatorTeamMessage( Level->Game->MessageMutator, this, P, PlayerReplicationInfo, Processed, TeamSayName, true) )
					P->eventTeamMessage( PlayerReplicationInfo, Processed, TeamSayName, true);
	}
}

//************************************
//Natives go here

/**
- Native functions are designed to retrieve parameters from the higher stack
- ProcessEvent creates a new stack with pre-stored locals and passes it 
  incorrectly because it's not designed to be used on native functions.
- Therefore variables don't come from code, but from pre-stored locals.
*/	

void APlayerPawn_Dummy::execTeamSay( FFrame& Stack, RESULT_DECL)
{
	guard(execTeamSay);

	//Classify our node's execution stack
	UFunction* F = Cast<UFunction>(Stack.Node);

	//Called via ProcessEvent >>> Native to Script
	if ( F && F->Func == (Native)&APlayerPawn_Dummy::execTeamSay )
	{
		FString*	MsgRef			= (FString*)	(Stack.Locals + 0);
		TeamSay( *MsgRef);
	}
	//Called via ProcessInternal >>> Script to Native
	else
	{
		P_GET_STR( Msg);	P_FINISH;
		TeamSay( Msg);
	}
	
	unguard;
}

void APlayerPawn_Dummy::execViewClass( FFrame& Stack, RESULT_DECL)
{
	guard(execViewclass);

	//Classify our node's execution stack
	UFunction* F = Cast<UFunction>(Stack.Node);
	
	//Called via ProcessEvent >>> Native to Script
	if ( F && F->Func == (Native)&APlayerPawn_Dummy::execViewClass )
	{
		UClass**	ActorClassRef	= (UClass**)	(Stack.Locals + 0);
		UBOOL*		bQuietRef		= (UBOOL*)		(Stack.Locals + 4);
		ViewClass( *ActorClassRef, *bQuietRef);
		return;
	}
	
	//Called via ProcessInternal >>> Script to Native
	P_GET_OBJECT( UClass, ActorClass);
	P_GET_UBOOL_OPTX( bQuiet, 0);
	P_FINISH;
	ViewClass( ActorClass, bQuiet);
	unguard;
}

void APlayerPawn_Dummy::execViewPlayer( FFrame& Stack, RESULT_DECL)
{
	guard(execViewPlayer);

	//Classify our node's execution stack
	UFunction* F = Cast<UFunction>(Stack.Node);
	
	//Called via ProcessEvent >>> Native to Script
	if ( F && F->Func == (Native)&APlayerPawn_Dummy::execViewPlayer )
	{
		FString*	SRef	= (FString*)	(Stack.Locals + 0);
		ViewPlayer( *SRef);
		return;
	}
	
	//Called via ProcessInternal >>> Script to Native
	P_GET_STR( S);
	P_FINISH;
	ViewPlayer( S);
	unguard;
}

void APlayerPawn_Dummy::execViewPlayerNum( FFrame& Stack, RESULT_DECL)
{
	guard(execViewPlayerNum);

	//Classify our node's execution stack and preallocate our only parameter (optx=0)
	UFunction* F = Cast<UFunction>(Stack.Node);
	INT num = 0;
	
	//Called via ProcessEvent >>> Native to Script
	if ( F && F->Func == (Native)&APlayerPawn_Dummy::execViewPlayerNum )
		num = *((INT*)Stack.Locals);
	//Called via ProcessInternal >>> Script to Native | and there's a variable to process
	else
	{
		Stack.Step( Stack.Object, &num); //if no variable is used, then Step does nothing (EX_EndFunctionParams)
		P_FINISH;
	}

	if ( !Level || !XLevel || !PlayerReplicationInfo )
		return;

	if ( !PlayerReplicationInfo->bIsSpectator && Level->Game && !Level->Game->bTeamGame )
		return;

	if ( num >= 0 )
	{
		APawn* P = Cast<APawn>(ViewTarget);
		if ( PlayerReplicationInfo->bIsSpectator ) //UTPure method for spectators!!
		{
			if ( P && P->bIsPlayer && P->PlayerReplicationInfo && (P->PlayerReplicationInfo->PlayerID == num) )
			{
				ViewTarget = NULL;
				bBehindView = 0;
				return;
			}
			for ( P=Level->PawnList; P; P=P->nextPawn )
				if ( P->PlayerReplicationInfo 
					&& !P->PlayerReplicationInfo->bIsSpectator
					&& (P->PlayerReplicationInfo->PlayerID == num) )
				{
					if ( P != this )
					{
						ViewTarget = P;
						bBehindView = 1;
					}
					break;
				}
			return;
		}
		if ( P && P->bIsPlayer && P->PlayerReplicationInfo && (P->PlayerReplicationInfo->TeamID == num) )
		{
			ViewTarget = NULL;
			bBehindView = 0;
			return;
		}
		for ( P=Level->PawnList; P; P=P->nextPawn )
			if ( P->PlayerReplicationInfo 
				&& (P->PlayerReplicationInfo->Team == PlayerReplicationInfo->Team)
				&& !P->PlayerReplicationInfo->bIsSpectator
				&& (P->PlayerReplicationInfo->TeamID == num) )
			{
				if ( P != this )
				{
					ViewTarget = P;
					bBehindView = 1;
				}
				break;
			}
		return;
	}

	if ( Role == ROLE_Authority )
	{
		if ( XCGE_Defaults->b451Setup && (ViewDelay > 0.0) )
			if (bCyclingView || ((Level->TimeSeconds - LastView) < ViewDelay))
				return;

		//This optimizes the following loop
		if ( ViewTarget && !ViewTarget->IsA(APawn::StaticClass() ) )
			ViewTarget = NULL;
	
		APawn* First = NULL;
		UBOOL bFound = 0;
		INT i = XLevel->iFirstDynamicActor; //Only dynamic actors
		for ( ; i<XLevel->Actors.Num() ; i++ )
		{
			APawn* P = Cast<APawn>(XLevel->Actors(i));
			if ( P && !P->bDeleteMe && P->PlayerReplicationInfo && !P->PlayerReplicationInfo->bIsSpectator )
			{
				if ( !First && (P != this) && ((bAdmin && !Level->Game) || GameCanSpectate(Level->Game, this, P) ) )
				{
					First = P;
					if ( (bFound == 2) || !ViewTarget ) //No need to keep querying
						break;
					bFound = 1;
				}
				if ( P == ViewTarget )
				{
					First = NULL;
					bFound = 2;
				}
			}
		}
		ViewTarget = First;
		
		if ( ViewTarget )
			eventClientMessage(ViewingFrom + GetBestName(ViewTarget), NAME_Event, 1);
		else
			eventClientMessage(ViewingFrom + OwnCamera, NAME_Event, 1);
		if ( XCGE_Defaults->b451Setup )
			LastView = Level->TimeSeconds;
	}
	unguard;
}